//-----------------------------------------------------------------------------
// Connection.c
//   Definition of the Python type OracleConnection.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// structure for the Python type "Connection"
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_HEAD
  OCISvcCtx *i_Handle;
  PyObject *i_UserName;
  PyObject *i_Password;
  PyObject *i_TNSEntry;
  PyObject *i_Version;
} udt_Connection;


//-----------------------------------------------------------------------------
// functions for the Python type "Connection"
//-----------------------------------------------------------------------------
static void Connection_Free(udt_Connection*);
static PyObject *Connection_GetAttr(udt_Connection*, PyObject*);
static PyObject *Connection_Close(udt_Connection*, PyObject*);
static PyObject *Connection_Commit(udt_Connection*, PyObject*);
static PyObject *Connection_Rollback(udt_Connection*, PyObject*);
static PyObject *Connection_NewCursor(udt_Connection*, PyObject*);


//-----------------------------------------------------------------------------
// declaration of Python type "Connection"
//-----------------------------------------------------------------------------
static PyTypeObject g_ConnectionType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "OracleConnection",			// type name string
  sizeof(udt_Connection),		// basic size (memory allocated)
  0,					// item size (variable object)
  (destructor) Connection_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Connection_GetAttr,	// getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// declaration of methods for Python type "Connection"
//-----------------------------------------------------------------------------
static PyMethodDef g_ConnectionMethods[] = {
  { "close", (PyCFunction) Connection_Close, METH_VARARGS },
  { "commit", (PyCFunction) Connection_Commit, METH_VARARGS },
  { "rollback", (PyCFunction) Connection_Rollback, METH_VARARGS },
  { "cursor", (PyCFunction) Connection_NewCursor, METH_VARARGS },
  { NULL, NULL }
};


//-----------------------------------------------------------------------------
// Connection_IsConnected()
//   Determines if the connection object is connected to the database. If not,
// a Python exception is raised.
//-----------------------------------------------------------------------------
static int Connection_IsConnected(
  udt_Connection *a_Connection)		// connection to check
{
  if (!a_Connection->i_Handle) {
    PyErr_SetString(g_InterfaceErrorException, "not connected");
    return -1;
  }
  return 0;
}


#include "Cursor.c"


//-----------------------------------------------------------------------------
// Connection_New()
//   Create a new connection object.
//-----------------------------------------------------------------------------
static PyObject *Connection_New(
  const char *a_UserName,		// user name
  unsigned a_UserNameLen,		// length of user name
  const char *a_Password,		// password
  unsigned a_PasswordLen,		// length of password
  const char *a_TNSEntry,		// TNS entry
  unsigned a_TNSEntryLen)		// length of TNS entry
{
  udt_Connection *v_Connection;
  sword v_Status;

  // create a new object for the Oracle connection
  v_Connection = PyObject_NEW(udt_Connection, &g_ConnectionType);
  if (!v_Connection)
    return NULL;
  v_Connection->i_Handle = NULL;
  v_Connection->i_UserName = NULL;
  v_Connection->i_Password = NULL;
  v_Connection->i_TNSEntry = NULL;
  v_Connection->i_Version = NULL;

  // create the string for the username
  v_Connection->i_UserName = PyString_FromStringAndSize(a_UserName,
      a_UserNameLen);
  if (!v_Connection->i_UserName) {
    Py_DECREF(v_Connection);
    return NULL;
  }

  // create the string for the password
  v_Connection->i_Password = PyString_FromStringAndSize(a_Password,
      a_PasswordLen);
  if (!v_Connection->i_Password) {
    Py_DECREF(v_Connection);
    return NULL;
  }

  // create the string for the TNS entry
  v_Connection->i_TNSEntry = PyString_FromStringAndSize(a_TNSEntry,
      a_TNSEntryLen);
  if (!v_Connection->i_TNSEntry) {
    Py_DECREF(v_Connection);
    return NULL;
  }

  // attempt logon to the database
  v_Status = OCILogon(
      g_EnvHandle,			// environment handle
      g_ErrorHandle,			// error handle
      &(v_Connection->i_Handle),	// service handle (OUT)
      (text*) a_UserName,		// username
      a_UserNameLen,			// username string length
      (text*) a_Password,		// password
      a_PasswordLen,			// password string length
      (text*) a_TNSEntry,		// TNS entry
      a_TNSEntryLen);			// TNS entry string length
  if (CheckForError(v_Status) < 0) {
    Py_DECREF(v_Connection);
    return NULL;
  }

  return (PyObject*) v_Connection;
}


//-----------------------------------------------------------------------------
// Connection_Free()
//   Deallocate the connection, disconnecting from the database if necessary.
//-----------------------------------------------------------------------------
static void Connection_Free(
  udt_Connection *a_Connection)		// connection object
{
  if (a_Connection->i_Handle) {
    OCILogoff(a_Connection->i_Handle, g_ErrorHandle);
    OCIHandleFree(a_Connection->i_Handle, OCI_HTYPE_SVCCTX);
  }
  Py_XDECREF(a_Connection->i_UserName);
  Py_XDECREF(a_Connection->i_Password);
  Py_XDECREF(a_Connection->i_TNSEntry);
  Py_XDECREF(a_Connection->i_Version);
  PyMem_DEL(a_Connection);
}


//-----------------------------------------------------------------------------
// Connection_Version()
//   Retrieve the version of the database and return it. Note that this
// function also places the result in the associated dictionary so it is only
// calculated once.
//-----------------------------------------------------------------------------
static PyObject *Connection_Version(
  udt_Connection *a_Connection)		// connection object
{
  PyObject *v_VersionVar, *v_CompatVar, *v_Temp, *v_Results;
  udt_Cursor *v_Cursor;
  int v_Count;

  // if version has already been determined, no need to determine again
  if (a_Connection->i_Version) {
    Py_INCREF(a_Connection->i_Version);
    return a_Connection->i_Version;
  }

  // allocate a cursor to retrieve the version
  v_Cursor = (udt_Cursor*) Cursor_New(a_Connection);
  if (!v_Cursor)
    return NULL;

  // set up cursor to determine if the version routine is available
  v_Temp = Py_BuildValue("(s)",
      "select count(*) from all_arguments where owner = 'SYS' " \
      "and package_name = 'DBMS_UTILITY' and object_name = 'DB_VERSION'");
  if (!v_Temp) {
    Py_DECREF(v_Cursor);
    return NULL;
  }

  // execute the cursor to determine if the version routine is available
  v_Results = Cursor_Execute(v_Cursor, v_Temp, NULL);
  Py_DECREF(v_Temp);
  if (!v_Results) {
    Py_DECREF(v_Cursor);
    return NULL;
  }

  // if not found, then simply assume 7.x and return
  Py_DECREF(v_Results);
  v_Results = Cursor_FetchOne(v_Cursor, NULL);
  if (!v_Results) {
    Py_DECREF(v_Cursor);
    return NULL;
  }
  v_Count = (int) PyFloat_AS_DOUBLE(PyTuple_GET_ITEM(v_Results, 0));
  Py_DECREF(v_Results);
  if (v_Count == 0) {
    Py_DECREF(v_Cursor);
    a_Connection->i_Version = PyString_FromString("7.x");
    if (a_Connection->i_Version)
      Py_INCREF(a_Connection->i_Version);
    return a_Connection->i_Version;
  }

  // create the bind variables
  v_VersionVar = StringVar_Bind(v_Cursor, 30, NULL, NULL);
  if (!v_VersionVar) {
    Py_DECREF(v_Cursor);
    return NULL;
  }
  v_CompatVar = StringVar_Bind(v_Cursor, 30, NULL, NULL);
  if (!v_CompatVar) {
    Py_DECREF(v_VersionVar);
    Py_DECREF(v_Cursor);
    return NULL;
  }
  v_Cursor->i_BindVariables = Py_BuildValue("{s:O,s:O}",
      "ver", v_VersionVar, "compat", v_CompatVar);
  Py_DECREF(v_CompatVar);
  if (!v_Cursor->i_BindVariables) {
    Py_DECREF(v_VersionVar);
    Py_DECREF(v_Cursor);
    return NULL;
  }

  // create the parameters for the function call
  v_Temp = Py_BuildValue("(s)",
      "begin dbms_utility.db_version(:ver, :compat); end;");
  if (!v_Temp) {
    Py_DECREF(v_VersionVar);
    Py_DECREF(v_Cursor);
    return NULL;
  }

  // execute the cursor
  v_Cursor->i_SetInputSizes = 1;
  v_Results = Cursor_Execute(v_Cursor, v_Temp, NULL);
  Py_DECREF(v_Temp);
  if (!v_Results) {
    Py_DECREF(v_Cursor);
    Py_DECREF(v_VersionVar);
    return NULL;
  }
  Py_DECREF(v_Results);

  // retrieve value
  a_Connection->i_Version = StringVar_Value((udt_StringVar*) v_VersionVar, 0);
  Py_DECREF(v_VersionVar);
  Py_DECREF(v_Cursor);
  if (!a_Connection->i_Version)
    return NULL;

  Py_INCREF(a_Connection->i_Version);
  return a_Connection->i_Version;
}


//-----------------------------------------------------------------------------
// Connection_GetAttr()
//   Retrieve an attribute on the connection object.
//-----------------------------------------------------------------------------
static PyObject *Connection_GetAttr(
  udt_Connection *a_Connection,		// connection object
  PyObject *a_Name)			// name of attribute
{
  PyObject *v_Attribute;
  char *v_Name;

  v_Name = PyString_AS_STRING(a_Name);
  v_Attribute = Py_FindMethod(g_ConnectionMethods, (PyObject*) a_Connection,
      v_Name);
  if (v_Attribute)
    return v_Attribute;
  PyErr_Clear();
  if (v_Name[0] == 'v' && strcmp(v_Name, "version") == 0)
    return Connection_Version(a_Connection);
  if (v_Name[0] == 'u' && strcmp(v_Name, "username") == 0) {
    Py_INCREF(a_Connection->i_UserName);
    return a_Connection->i_UserName;
  } else if (v_Name[0] == 'p' && strcmp(v_Name, "password") == 0) {
    Py_INCREF(a_Connection->i_Password);
    return a_Connection->i_Password;
  } else if (v_Name[0] == 't' && strcmp(v_Name, "tnsentry") == 0) {
    Py_INCREF(a_Connection->i_TNSEntry);
    return a_Connection->i_TNSEntry;
  }
  PyErr_SetObject(PyExc_AttributeError, a_Name);
  return NULL;
}


//-----------------------------------------------------------------------------
// Connection_Close()
//   Disconnect from the connection.
//-----------------------------------------------------------------------------
static PyObject *Connection_Close(
  udt_Connection *a_Connection,		// connection to close
  PyObject *a_Args)			// arguments
{
  sword v_Status;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure we are actually connected
  if (Connection_IsConnected(a_Connection) < 0)
    return NULL;

  // logoff of the server
  v_Status = OCILogoff(a_Connection->i_Handle, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;
  OCIHandleFree(a_Connection->i_Handle, OCI_HTYPE_SVCCTX);
  a_Connection->i_Handle = NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Connection_Commit()
//   Commit the transaction on the connection.
//-----------------------------------------------------------------------------
static PyObject *Connection_Commit(
  udt_Connection *a_Connection,		// connection to close
  PyObject *a_Args)			// arguments
{
  sword v_Status;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure we are actually connected
  if (Connection_IsConnected(a_Connection) < 0)
    return NULL;

  // perform the commit
  v_Status = OCITransCommit(
      a_Connection->i_Handle,		// service handle
      g_ErrorHandle,			// error handle
      OCI_DEFAULT);			// mode
  if (CheckForError(v_Status) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Connection_Rollback()
//   Rollback the transaction on the connection.
//-----------------------------------------------------------------------------
static PyObject *Connection_Rollback(
  udt_Connection *a_Connection,		// connection to close
  PyObject *a_Args)			// arguments
{
  sword v_Status;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure we are actually connected
  if (Connection_IsConnected(a_Connection) < 0)
    return NULL;

  // perform the rollback
  v_Status = OCITransRollback(
      a_Connection->i_Handle,		// service handle
      g_ErrorHandle,			// error handle
      OCI_DEFAULT);			// mode
  if (CheckForError(v_Status) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Connection_NewCursor()
//   Create a new cursor (statement) referencing the connection.
//-----------------------------------------------------------------------------
static PyObject *Connection_NewCursor(
  udt_Connection *a_Connection,		// connection to close
  PyObject *a_Args)			// arguments
{
  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  return Cursor_New(a_Connection);
}

