//-----------------------------------------------------------------------------
// Cursor.c
//   Definition of the Python type OracleCursor.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// structure for the Python type "Cursor"
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_HEAD
  OCIStmt *i_Handle;
  udt_Connection *i_Connection;
  PyObject *i_BindVariables;
  PyObject *i_DefineVariables;
  int i_ArraySize;
  int i_SetInputSizes;
  int i_RowCount;
  int i_AllocatedRows;
  int i_ActualRows;
  int i_RowNum;
} udt_Cursor;


#include "Variable.c"


//-----------------------------------------------------------------------------
// dependent function defintions
//-----------------------------------------------------------------------------
static void Cursor_Free(udt_Cursor*);


//-----------------------------------------------------------------------------
// functions for the Python type "Cursor"
//-----------------------------------------------------------------------------
static PyObject *Cursor_GetAttr(udt_Cursor*, PyObject*);
static int Cursor_SetAttr(udt_Cursor*, PyObject*, PyObject*);
static PyObject *Cursor_Close(udt_Cursor*, PyObject*);
static PyObject *Cursor_Execute(udt_Cursor*, PyObject*, PyObject*);
static PyObject *Cursor_FetchOne(udt_Cursor*, PyObject*);
static PyObject *Cursor_FetchMany(udt_Cursor*, PyObject*);
static PyObject *Cursor_FetchAll(udt_Cursor*, PyObject*);
static PyObject *Cursor_FetchBinds(udt_Cursor*, PyObject*);
static PyObject *Cursor_SetInputSizes(udt_Cursor*, PyObject*, PyObject*);


//-----------------------------------------------------------------------------
// declaration of Python type "Cursor"
//-----------------------------------------------------------------------------
static PyTypeObject g_CursorType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "OracleCursor",			// type name string
  sizeof(udt_Cursor),			// basic size (memory allocated)
  0,					// item size (variable object)
  (destructor) Cursor_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Cursor_GetAttr,	// getattro function
  (setattrofunc) Cursor_SetAttr,	// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// declaration of methods for Python type "Cursor"
//-----------------------------------------------------------------------------
static PyMethodDef g_CursorMethods[] = {
  { "close", (PyCFunction) Cursor_Close, METH_VARARGS },
  { "execute", (PyCFunction) Cursor_Execute, METH_VARARGS | METH_KEYWORDS },
  { "fetchone", (PyCFunction) Cursor_FetchOne, METH_VARARGS },
  { "fetchmany", (PyCFunction) Cursor_FetchMany, METH_VARARGS },
  { "fetchall", (PyCFunction) Cursor_FetchAll, METH_VARARGS },
  { "fetchbinds", (PyCFunction) Cursor_FetchBinds, METH_VARARGS },
  { "setinputsizes", (PyCFunction) Cursor_SetInputSizes,
        METH_VARARGS | METH_KEYWORDS },
  { NULL, NULL }
};


//-----------------------------------------------------------------------------
// Cursor_IsOpen()
//   Determines if the cursor object is open and if so, if the connection is
// also open.
//-----------------------------------------------------------------------------
static int Cursor_IsOpen(
  udt_Cursor *a_Cursor)			// cursor to check
{
  if (!a_Cursor->i_Handle) {
    PyErr_SetString(g_InterfaceErrorException, "not open");
    return -1;
  }
  return Connection_IsConnected(a_Cursor->i_Connection);
}


//-----------------------------------------------------------------------------
// Cursor_New()
//   Create a new cursor object.
//-----------------------------------------------------------------------------
static PyObject *Cursor_New(
  udt_Connection *a_Connection)		// connection to associate with
{
  udt_Cursor *v_Cursor;
  sword v_Status;

  // create a new object for the Oracle connection
  v_Cursor = PyObject_NEW(udt_Cursor, &g_CursorType);
  if (!v_Cursor)
    return NULL;
  v_Cursor->i_Handle = NULL;
  Py_INCREF(a_Connection);
  v_Cursor->i_Connection = a_Connection;
  v_Cursor->i_BindVariables = NULL;
  v_Cursor->i_DefineVariables = NULL;
  v_Cursor->i_ArraySize = 1;
  v_Cursor->i_RowCount = 0;
  v_Cursor->i_SetInputSizes = 0;

  // allocate an Oracle statement handle
  v_Status = OCIHandleAlloc(g_EnvHandle, (dvoid**) &v_Cursor->i_Handle,
      OCI_HTYPE_STMT, 0, 0);
  if (CheckForError(v_Status) < 0) {
    Py_DECREF(v_Cursor);
    return NULL;
  }

  return (PyObject*) v_Cursor;
}


//-----------------------------------------------------------------------------
// Cursor_Free()
//   Deallocate the cursor.
//-----------------------------------------------------------------------------
static void Cursor_Free(
  udt_Cursor *a_Cursor)			// cursor object
{
  Py_DECREF(a_Cursor->i_Connection);
  if (a_Cursor->i_Handle)
    OCIHandleFree(a_Cursor->i_Handle, OCI_HTYPE_STMT);
  Py_XDECREF(a_Cursor->i_BindVariables);
  Py_XDECREF(a_Cursor->i_DefineVariables);
  PyMem_DEL(a_Cursor);
}


//-----------------------------------------------------------------------------
// Cursor_ItemDescription()
//   Return a tuple describing the item at the given position.
//-----------------------------------------------------------------------------
static PyObject *Cursor_ItemDescription(
  udt_Cursor *a_Cursor,			// cursor object
  unsigned a_Pos)			// position
{
  int v_DisplaySize, v_Index;
  PyObject *v_Tuple, *v_Var;
  ub2 v_InternalSize;
  ub1 v_Precision;
  dvoid *v_Param;
  sword v_Status;
  ub4 v_NameLen;
  char *v_Name;
  sb1 v_Scale;

  // acquire parameter descriptor
  v_Status = OCIParamGet(a_Cursor->i_Handle, OCI_HTYPE_STMT, g_ErrorHandle,
      &v_Param, a_Pos);
  if (CheckForError(v_Status) < 0)
    return NULL;

  // acquire internal size of item
  v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_InternalSize,
      0, OCI_ATTR_DATA_SIZE, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;

  // aquire name of item
  v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_Name,
      &v_NameLen, OCI_ATTR_NAME, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;

  // lookup precision and scale
  v_Var = PyList_GET_ITEM(a_Cursor->i_DefineVariables, a_Pos - 1);
  if (v_Var->ob_type == &g_NumberVarType) {
    v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_Scale,
        0, OCI_ATTR_SCALE, g_ErrorHandle);
    if (CheckForError(v_Status) < 0)
      return NULL;
    v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_Precision,
        0, OCI_ATTR_PRECISION, g_ErrorHandle);
    if (CheckForError(v_Status) < 0)
      return NULL;
  }

  // set display length based on data type
  if (v_Var->ob_type == &g_StringVarType)
    v_DisplaySize = v_InternalSize;
  else if (v_Var->ob_type == &g_NumberVarType) {
    v_DisplaySize = v_Precision;
    if (v_Scale)
      v_DisplaySize += v_Scale + 1;
  } else if (v_Var->ob_type == &g_DateTimeVarType) {
    v_DisplaySize = 23;
  } else {
    v_DisplaySize = -1;
  }

  // create the tuple and populate it
  v_Tuple = PyTuple_New(7);
  if (!v_Tuple)
    return NULL;

  // set each of the items in the tuple
  PyTuple_SET_ITEM(v_Tuple, 0, PyString_FromStringAndSize(v_Name, v_NameLen));
  Py_INCREF(v_Var->ob_type);
  PyTuple_SET_ITEM(v_Tuple, 1, (PyObject*) v_Var->ob_type);
  PyTuple_SET_ITEM(v_Tuple, 2, PyInt_FromLong(v_DisplaySize));
  PyTuple_SET_ITEM(v_Tuple, 3, PyInt_FromLong(v_InternalSize));
  PyTuple_SET_ITEM(v_Tuple, 4, PyInt_FromLong(v_Precision));
  PyTuple_SET_ITEM(v_Tuple, 5, PyInt_FromLong(v_Scale));
  PyTuple_SET_ITEM(v_Tuple, 6, PyInt_FromLong(1));

  // make sure the tuple is ok
  for (v_Index = 0; v_Index < 7; v_Index++) {
    if (!PyTuple_GET_ITEM(v_Tuple, v_Index)) {
      Py_DECREF(v_Tuple);
      return NULL;
    }
  }

  return v_Tuple;
}


//-----------------------------------------------------------------------------
// Cursor_Description()
//   Return a list of 7-tuples consisting of the description of the define
// variables.
//-----------------------------------------------------------------------------
static PyObject *Cursor_Description(
  udt_Cursor *a_Cursor)			// cursor object
{
  PyObject *v_Results, *v_Tuple;
  int v_NumItems, v_Index;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // if the cursor is not a query, return None
  if (!a_Cursor->i_DefineVariables) {
    Py_INCREF(Py_None);
    return Py_None;
  }

  // create a list of the required length
  v_NumItems = PyList_GET_SIZE(a_Cursor->i_DefineVariables);
  v_Results = PyList_New(v_NumItems);
  if (!v_Results)
    return NULL;

  // create tuples corresponding to the select-items
  for (v_Index = 0; v_Index < v_NumItems; v_Index++) {
    v_Tuple = Cursor_ItemDescription(a_Cursor, v_Index + 1);
    if (!v_Tuple) {
      Py_DECREF(v_Results);
      return NULL;
    }
    PyList_SET_ITEM(v_Results, v_Index, v_Tuple);
  }

  return v_Results;
}


//-----------------------------------------------------------------------------
// Cursor_GetAttr()
//   Retrieve an attribute on the connection object.
//-----------------------------------------------------------------------------
static PyObject *Cursor_GetAttr(
  udt_Cursor *a_Cursor,			// cursor object
  PyObject *a_Name)			// name of attribute
{
  PyObject *v_Attribute;
  char *v_Name;

  v_Name = PyString_AS_STRING(a_Name);
  v_Attribute = Py_FindMethod(g_CursorMethods, (PyObject*) a_Cursor, v_Name);
  if (v_Attribute)
    return v_Attribute;
  PyErr_Clear();
  if (v_Name[0] == 'r' && strcmp(v_Name, "rowcount") == 0)
    return PyInt_FromLong(a_Cursor->i_RowCount);
  if (v_Name[0] == 'a' && strcmp(v_Name, "arraysize") == 0)
    return PyInt_FromLong(a_Cursor->i_ArraySize);
  if (v_Name[0] == 'd' && strcmp(v_Name, "description") == 0)
    return Cursor_Description(a_Cursor);
  PyErr_SetObject(PyExc_AttributeError, a_Name);
  return NULL;
}


//-----------------------------------------------------------------------------
// Cursor_SetAttr()
//   Retrieve an attribute on the connection object.
//-----------------------------------------------------------------------------
static int Cursor_SetAttr(
  udt_Cursor *a_Cursor,			// cursor object
  PyObject *a_Name,			// name of attribute
  PyObject *a_Value)			// value to set
{
  const char *v_Name;

  v_Name = PyString_AS_STRING(a_Name);
  if (v_Name[0] == 'a' && strcmp(v_Name, "arraysize") == 0) {
    if (!PyInt_Check(a_Value)) {
      PyErr_SetString(PyExc_TypeError, "array size must be an integer");
      return -1;
    }
    a_Cursor->i_ArraySize = PyInt_AS_LONG(a_Value);
    return 0;
  }
  PyErr_SetObject(PyExc_AttributeError, a_Name);
  return -1;
}


//-----------------------------------------------------------------------------
// Cursor_Close()
//   Close the cursor.
//-----------------------------------------------------------------------------
static PyObject *Cursor_Close(
  udt_Cursor *a_Cursor,			// cursor to close
  PyObject *a_Args)			// arguments
{
  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure we are actually open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // close the cursor
  OCIHandleFree(a_Cursor->i_Handle, OCI_HTYPE_STMT);
  a_Cursor->i_Handle = NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_PerformBind()
//   Perform the binds on the cursor.
//-----------------------------------------------------------------------------
static int Cursor_PerformBind(
  udt_Cursor *a_Cursor,			// cursor to perform binds on
  PyObject *a_Parameters)		// parameters to bind
{
  PyObject *v_Key, *v_Value, *v_Var;
  int v_Pos;

  // clear bind variable dictionary, unless previously initialized
  if (!a_Cursor->i_SetInputSizes) {
    Py_XDECREF(a_Cursor->i_BindVariables);
    a_Cursor->i_BindVariables = PyDict_New();
    if (!a_Cursor->i_BindVariables)
      return -1;
  }

  // set values and perform binds for all out or in/out variables
  if (a_Cursor->i_SetInputSizes) {
    v_Pos = 0;
    while (PyDict_Next(a_Cursor->i_BindVariables, &v_Pos, &v_Key, &v_Var)) {
      if (a_Parameters) {
        v_Value = PyDict_GetItem(a_Parameters, v_Key);
        if (v_Value && Variable_SetValue(v_Var, v_Value) < 0)
          return -1;
      }
      if (Variable_BindToCursor(a_Cursor, v_Key, v_Var) < 0)
        return -1;
    }
  }

  // perform the remaining binds
  if (a_Parameters) {
    v_Pos = 0;
    while (PyDict_Next(a_Parameters, &v_Pos, &v_Key, &v_Value)) {
      v_Var = NULL;
      if (a_Cursor->i_SetInputSizes)
        v_Var = PyDict_GetItem(a_Cursor->i_BindVariables, v_Key);
      if (!v_Var) {
        v_Var = Variable_Bind(a_Cursor, v_Key, v_Value);
        if (!v_Var)
          return -1;
        if (PyDict_SetItem(a_Cursor->i_BindVariables, v_Key, v_Var) < 0) {
          Py_DECREF(v_Var);
          return -1;
        }
        Py_DECREF(v_Var);
      }
    }
  }

  // ensure that input sizes are reset
  a_Cursor->i_SetInputSizes = 0;

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_PerformDefine()
//   Perform the defines for the cursor. At this point it is assumed that the
// statement being executed is in fact a query.
//-----------------------------------------------------------------------------
static int Cursor_PerformDefine(
  udt_Cursor *a_Cursor,			// cursor to perform binds on
  int a_IsQuery)			// is a query statement?
{
  PyObject *v_Var;
  int v_NumParams, v_Pos;
  sword v_Status;

  // if not a query, simply replace the list of define variables with None
  if (!a_IsQuery) {
    if (a_Cursor->i_DefineVariables) {
      Py_DECREF(a_Cursor->i_DefineVariables);
      a_Cursor->i_DefineVariables = NULL;
    }
    return 0;
  }

  // ensure the state parameters are correct
  a_Cursor->i_RowCount = 0;
  a_Cursor->i_ActualRows = -1;
  a_Cursor->i_RowNum = 0;

  // determine number of items in select-list
  v_Status = OCIAttrGet(a_Cursor->i_Handle, OCI_HTYPE_STMT,
      (dvoid*) &v_NumParams, 0, OCI_ATTR_PARAM_COUNT, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return -1;

  // create a list corresponding to the number of items
  Py_XDECREF(a_Cursor->i_DefineVariables);
  a_Cursor->i_DefineVariables = PyList_New(v_NumParams);
  if (!a_Cursor->i_DefineVariables)
    return -1;

  // define a variable for each select-item
  for (v_Pos = 1; v_Pos <= v_NumParams; v_Pos++) {
    v_Var = Variable_Define(a_Cursor, v_Pos);
    if (!v_Var)
      return -1;
    PyList_SET_ITEM(a_Cursor->i_DefineVariables, v_Pos - 1, v_Var);
  }

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_CreateTuple()
//   Create a tuple consisting of each of the items in the select-list.
//-----------------------------------------------------------------------------
static PyObject *Cursor_CreateTuple(
  udt_Cursor *a_Cursor)			// cursor object
{
  PyObject *v_Tuple, *v_Item;
  int v_NumItems, v_Pos;

  // create a new tuple
  v_NumItems = PyList_GET_SIZE(a_Cursor->i_DefineVariables);
  v_Tuple = PyTuple_New(v_NumItems);
  if (!v_Tuple)
    return NULL;

  // acquire the value for each item
  for (v_Pos = 0; v_Pos < v_NumItems; v_Pos++) {
    v_Item = Variable_Value(PyList_GET_ITEM(a_Cursor->i_DefineVariables,
        v_Pos), a_Cursor->i_RowNum);
    if (!v_Item) {
      Py_DECREF(v_Tuple);
      return NULL;
    }
    PyTuple_SET_ITEM(v_Tuple, v_Pos, v_Item);
  }

  a_Cursor->i_RowNum++;
  a_Cursor->i_RowCount++;
  return v_Tuple;
}


//-----------------------------------------------------------------------------
// Cursor_Execute()
//   Commit the transaction on the connection.
//-----------------------------------------------------------------------------
static PyObject *Cursor_Execute(
  udt_Cursor *a_Cursor,			// cursor to execute
  PyObject *a_Args,			// arguments
  PyObject *a_Keywords)			// keywords
{
  unsigned v_IsQuery, v_StatementLen;
  const char *v_Statement;
  ub2 v_StatementType;
  ub4 v_RowCount;
  sword v_Status;

  // statement text is expected
  if (!PyArg_ParseTuple(a_Args, "s#", &v_Statement, &v_StatementLen))
    return NULL;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // determine the number of rows for allocation
  a_Cursor->i_AllocatedRows = ((a_Cursor->i_ArraySize >> 2) + 1) << 2;

  // prepare statement
  v_Status = OCIStmtPrepare(a_Cursor->i_Handle, g_ErrorHandle,
      (text*) v_Statement, v_StatementLen, OCI_NTV_SYNTAX, OCI_DEFAULT);
  if (CheckForError(v_Status) < 0)
    return NULL;

  // determine type of statement
  v_Status = OCIAttrGet(a_Cursor->i_Handle, OCI_HTYPE_STMT,
      (dvoid*) &v_StatementType, 0, OCI_ATTR_STMT_TYPE, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;
  v_IsQuery = (v_StatementType == OCI_STMT_SELECT);

  // perform binds
  if (Cursor_PerformBind(a_Cursor, a_Keywords) < 0)
    return NULL;

  // execute the statement
  v_Status = OCIStmtExecute(a_Cursor->i_Connection->i_Handle,
      a_Cursor->i_Handle, g_ErrorHandle, v_IsQuery ? 0 : 1, 0, 0, 0,
      OCI_DEFAULT);
  if (CheckForError(v_Status) < 0)
    return NULL;

  // set the rowcount attribute
  v_Status = OCIAttrGet(a_Cursor->i_Handle, OCI_HTYPE_STMT,
      &v_RowCount, 0, OCI_ATTR_ROW_COUNT, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;
  a_Cursor->i_RowCount = v_RowCount;

  // perform defines
  if (Cursor_PerformDefine(a_Cursor, v_IsQuery) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_VerifyFetch()
//   Fetch a single row from the cursor.
//-----------------------------------------------------------------------------
static int Cursor_VerifyFetch(
  udt_Cursor *a_Cursor)			// cursor to fetch from
{
  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return -1;

  // make sure the cursor is for a query
  if (!a_Cursor->i_DefineVariables) {
    PyErr_SetString(g_InterfaceErrorException, "not a query");
    return -1;
  }

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_MoreRows()
//   Returns an integer indicating if more rows can be retrieved from the
// cursor.
//-----------------------------------------------------------------------------
static int Cursor_MoreRows(
  udt_Cursor *a_Cursor)			// cursor to fetch from
{
  sword v_Status;
  ub4 v_RowCount;

  if (a_Cursor->i_RowNum >= a_Cursor->i_ActualRows) {
    if (a_Cursor->i_ActualRows < 0 ||
        a_Cursor->i_ActualRows == a_Cursor->i_AllocatedRows) {
      v_Status = OCIStmtFetch(a_Cursor->i_Handle, g_ErrorHandle,
          a_Cursor->i_AllocatedRows, OCI_FETCH_NEXT, OCI_DEFAULT);
      if (v_Status != OCI_NO_DATA && CheckForError(v_Status) < 0)
        return -1;
      v_Status = OCIAttrGet(a_Cursor->i_Handle, OCI_HTYPE_STMT,
          &v_RowCount, 0, OCI_ATTR_ROW_COUNT, g_ErrorHandle);
      if (CheckForError(v_Status) < 0)
        return -1;
      a_Cursor->i_ActualRows = v_RowCount - a_Cursor->i_RowCount;
      a_Cursor->i_RowNum = 0;
    }
    if (a_Cursor->i_RowNum >= a_Cursor->i_ActualRows)
      return 0;
  }

  return 1;
}


//-----------------------------------------------------------------------------
// Cursor_MultiFetch()
//   Return a list consisting of the remaining rows up to the given row limit
// (if specified).
//-----------------------------------------------------------------------------
static PyObject *Cursor_MultiFetch(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  int a_RowLimit)			// row limit
{
  PyObject *v_Results, *v_Tuple;
  int v_Row, v_Status;

  // create an empty list
  v_Results = PyList_New(0);
  if (!v_Results)
    return NULL;

  // fetch as many rows as possible
  for (v_Row = 0; a_RowLimit == 0 || v_Row < a_RowLimit; v_Row++) {
    v_Status = Cursor_MoreRows(a_Cursor);
    if (v_Status < 0) {
      Py_DECREF(v_Results);
      return NULL;
    } else if (v_Status == 0) {
      break;
    } else {
      v_Tuple = Cursor_CreateTuple(a_Cursor);
      if (!v_Tuple) {
        Py_DECREF(v_Results);
        return NULL;
      }
      if (PyList_Append(v_Results, v_Tuple) < 0) {
        Py_DECREF(v_Tuple);
        Py_DECREF(v_Results);
        return NULL;
      }
      Py_DECREF(v_Tuple);
    }
  }

  return v_Results;
}


//-----------------------------------------------------------------------------
// Cursor_FetchOne()
//   Fetch a single row from the cursor.
//-----------------------------------------------------------------------------
static PyObject *Cursor_FetchOne(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  int v_Status;

  // no arguments are expected
  if (a_Args && !PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // verify fetch can be performed
  if (Cursor_VerifyFetch(a_Cursor) < 0)
    return NULL;

  // setup return value
  v_Status = Cursor_MoreRows(a_Cursor);
  if (v_Status < 0)
    return NULL;
  else if (v_Status > 0)
    return Cursor_CreateTuple(a_Cursor);

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_FetchMany()
//   Fetch multiple rows from the cursor based on the arraysize.
//-----------------------------------------------------------------------------
static PyObject *Cursor_FetchMany(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  int v_RowLimit;

  // parse arguments -- optional rowlimit expected
  v_RowLimit = a_Cursor->i_ArraySize;
  if (!PyArg_ParseTuple(a_Args, "|i", &v_RowLimit))
    return NULL;

  // verify fetch can be performed
  if (Cursor_VerifyFetch(a_Cursor) < 0)
    return NULL;

  return Cursor_MultiFetch(a_Cursor, v_RowLimit);
}


//-----------------------------------------------------------------------------
// Cursor_FetchAll()
//   Fetch multiple rows from the cursor based on the arraysize.
//-----------------------------------------------------------------------------
static PyObject *Cursor_FetchAll(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // verify fetch can be performed
  if (Cursor_VerifyFetch(a_Cursor) < 0)
    return NULL;

  return Cursor_MultiFetch(a_Cursor, 0);
}

//-----------------------------------------------------------------------------
// Cursor_FetchBinds()
//   Fetch bind variables and their values (useful after an execute statment
// has occurred).
//-----------------------------------------------------------------------------
static PyObject *Cursor_FetchBinds(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  PyObject *v_Dict, *v_Key, *v_Var, *v_Value;
  int v_Pos;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // create an empty dictionary
  v_Dict = PyDict_New();
  if (!v_Dict)
    return NULL;

  // process each bind variable
  v_Pos = 0;
  while (PyDict_Next(a_Cursor->i_BindVariables, &v_Pos, &v_Key, &v_Var)) {
    v_Value = Variable_Value(v_Var, 0);
    if (!v_Value) {
      Py_DECREF(v_Dict);
      return NULL;
    }
    if (PyDict_SetItem(v_Dict, v_Key, v_Value) < 0) {
      Py_DECREF(v_Dict);
      Py_DECREF(v_Value);
      return NULL;
    }
    Py_DECREF(v_Value);
  }

  return v_Dict;
}

//-----------------------------------------------------------------------------
// Cursor_SetInputSizes()
//   Set the sizes of the bind variables.
//-----------------------------------------------------------------------------
static PyObject *Cursor_SetInputSizes(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args,			// arguments
  PyObject *a_Keywords)			// keyword arguments
{
  PyObject *v_Key, *v_Value, *v_Var;
  int v_Pos;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // eliminate existing bind variables
  Py_XDECREF(a_Cursor->i_BindVariables);
  a_Cursor->i_BindVariables = PyDict_New();
  if (!a_Cursor->i_BindVariables)
    return NULL;
  a_Cursor->i_SetInputSizes = 1;

  // process each input
  if (a_Keywords) {
    v_Pos = 0;
    while (PyDict_Next(a_Keywords, &v_Pos, &v_Key, &v_Value)) {
      v_Var = Variable_BindNoValue(a_Cursor, v_Value);
      if (!v_Var)
        return NULL;
      if (PyDict_SetItem(a_Cursor->i_BindVariables, v_Key, v_Var) < 0) {
        Py_DECREF(v_Var);
        return NULL;
      }
      Py_DECREF(v_Var);
    }
  }

  Py_INCREF(Py_None);
  return Py_None;
}

