//-----------------------------------------------------------------------------
// DateTimeVar.c
//   Defines the routines for handling numeric variables.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// DateTime type
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_VAR_HEAD
  unsigned i_MaxRows;
  sb2 *i_Indicator;
  OCIDate *i_Data;
} udt_DateTimeVar;

#include "ExternalDateTimeVar.c"

//-----------------------------------------------------------------------------
// Declaration of date/time variable functions.
//-----------------------------------------------------------------------------
static void DateTimeVar_Free(udt_DateTimeVar*);


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_DateTimeVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "DateTimeVar",			// type name string
  sizeof(udt_DateTimeVar),		// basic size (memory allocated)
  sizeof(OCIDate) + sizeof(sb2),	// item size (variable object)
  (destructor) DateTimeVar_Free,	// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  0,					// getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// DateTimeVar_New()
//   Create a new date/time variable.
//-----------------------------------------------------------------------------
PyObject *DateTimeVar_New(
  unsigned a_NumRows)			// number of rows to allocate
{
  udt_DateTimeVar *v_Var;

  // create Python object
  v_Var = PyObject_NEW_VAR(udt_DateTimeVar, &g_DateTimeVarType, a_NumRows);
  if (!v_Var)
    return NULL;

  // initialize members
  v_Var->i_MaxRows = a_NumRows;
  v_Var->i_Indicator = (sb2*) (((char*) v_Var) + sizeof(udt_DateTimeVar));
  v_Var->i_Data = (OCIDate*) &(v_Var->i_Indicator[a_NumRows]);

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// DateTimeVar_Free()
//   Free a date/time variable.
//-----------------------------------------------------------------------------
static void DateTimeVar_Free(
  udt_DateTimeVar *a_Var)		// variable to free
{
  PyMem_DEL(a_Var);
}


//-----------------------------------------------------------------------------
// DateTimeVar_Define()
//   Define a date/time variable at the specified position in the cursor.
//-----------------------------------------------------------------------------
PyObject *DateTimeVar_Define(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  int a_Position,			// position at which to perform define
  OCIParam *a_Param)			// parameter descriptor
{
  OCIDefine *v_Handle;
  udt_DateTimeVar *v_Var;
  sword v_Status;

  // create a date/time variable
  v_Var = (udt_DateTimeVar*) DateTimeVar_New(a_Cursor->i_AllocatedRows);
  if (!v_Var)
    return NULL;

  // perform the define
  v_Handle = 0;
  v_Status = OCIDefineByPos(a_Cursor->i_Handle, &v_Handle, g_ErrorHandle,
      a_Position, v_Var->i_Data, sizeof(OCIDate),  SQLT_ODT,
      v_Var->i_Indicator, 0, 0, OCI_DEFAULT);
  if (CheckForError(v_Status) < 0) {
    Py_DECREF(v_Var);
    return NULL;
  }

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// DateTimeVar_BindToCursor()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int DateTimeVar_BindToCursor(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  udt_DateTimeVar *a_Var)		// variable to bind
{
  OCIBind *v_BindHandle;
  sword v_Status;

  v_BindHandle = NULL;
  v_Status = OCIBindByName(a_Cursor->i_Handle, &v_BindHandle, g_ErrorHandle,
      PyString_AS_STRING(a_Name), PyString_GET_SIZE(a_Name), a_Var->i_Data,
      sizeof(OCIDate), SQLT_ODT, a_Var->i_Indicator, 0, 0, 0, 0,
      OCI_DEFAULT);
  if (CheckForError(v_Status) < 0)
    return -1;

  return 0;
}


//-----------------------------------------------------------------------------
// DateTimeVar_SetValue()
//   Set the value of the date.
//-----------------------------------------------------------------------------
static int DateTimeVar_SetValue(
  udt_DateTimeVar *a_Var,		// variable to set
  int a_Pos,				// position to set
  PyObject *a_Value)			// value to set
{
  udt_ExternalDateTimeVar *v_Value;
  sword v_Status;
  uword v_Valid;

  if (a_Value == Py_None)
    a_Var->i_Indicator[a_Pos] = OCI_IND_NULL;
  else {
    a_Var->i_Indicator[a_Pos] = OCI_IND_NOTNULL;
    v_Value = (udt_ExternalDateTimeVar*) a_Value;
    OCIDateSetDate(&a_Var->i_Data[a_Pos], v_Value->i_Year, v_Value->i_Month,
        v_Value->i_Day);
    OCIDateSetTime(&a_Var->i_Data[a_Pos], v_Value->i_Hour, v_Value->i_Minute,
        v_Value->i_Second);
    v_Status = OCIDateCheck(g_ErrorHandle, &a_Var->i_Data[a_Pos], &v_Valid);
    if (CheckForError(v_Status) < 0)
      return -1;
    if (v_Valid != 0) {
      PyErr_SetString(g_DataErrorException, "invalid date");
      return -1;
    }
  }
  return 0;
}


//-----------------------------------------------------------------------------
// DateTimeVar_Bind()
//   Bind a date to the variable name.
//-----------------------------------------------------------------------------
PyObject *DateTimeVar_Bind(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  PyObject *a_Value)			// value (null or integer/float types)
{
  udt_DateTimeVar *v_Var;

  v_Var = (udt_DateTimeVar*) DateTimeVar_New(1);
  if (!v_Var)
    return NULL;
  if (a_Value) {
    if (DateTimeVar_SetValue(v_Var, 0, a_Value) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
    if (DateTimeVar_BindToCursor(a_Cursor, a_Name, v_Var) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
  } else {
    v_Var->i_Indicator[0] = OCI_IND_NULL;
  }
  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// DateTimeVar_Value()
//   Returns the value for a date/time variable.
//-----------------------------------------------------------------------------
PyObject *DateTimeVar_Value(
  udt_DateTimeVar *a_Var,		// variable to determine value for
  int a_ArrayPos)			// array position
{
  if (a_Var->i_Indicator[a_ArrayPos] == OCI_IND_NULL) {
    Py_INCREF(Py_None);
    return Py_None;
  }
  return ExternalDateTimeVar_NewFromOracle(a_Var, a_ArrayPos);
}


//-----------------------------------------------------------------------------
// DateTimeVar_BindArrayToCursor()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int DateTimeVar_BindArrayToCursor(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  udt_DateTimeVar *a_Var)		// variable to bind
{
  OCIBind *v_BindHandle;
  sword v_Status;

  v_BindHandle = NULL;
  v_Status = OCIBindByName(a_Cursor->i_Handle, &v_BindHandle, g_ErrorHandle,
      PyString_AS_STRING(a_Name), PyString_GET_SIZE(a_Name), a_Var->i_Data,
      sizeof(OCIDate), SQLT_ODT, a_Var->i_Indicator, 0, 0, a_Var->i_MaxRows,
      &(a_Var->i_MaxRows), OCI_DEFAULT);
  if (CheckForError(v_Status) < 0)
    return -1;

  return 0;
}

//-----------------------------------------------------------------------------
// DateTimeVar_BindArray()
//   Bind a date to the variable name.
//-----------------------------------------------------------------------------
PyObject *DateTimeVar_BindArray(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  PyObject *a_Value)			// value (null or integer/float types)
{
  udt_DateTimeVar *v_Var;
  int i, v_Size;

  v_Size = PyList_GET_SIZE(a_Value);
  v_Var = (udt_DateTimeVar*) DateTimeVar_New(v_Size);
  if (!v_Var)
    return NULL;
  if (a_Value) {
    for (i = 0; i < v_Size; i++) {
      if (DateTimeVar_SetValue(v_Var, i, PyList_GET_ITEM(a_Value, i)) < 0) {
        Py_DECREF(v_Var);
        return NULL;
      }
    }
    if (DateTimeVar_BindArrayToCursor(a_Cursor, a_Name, v_Var) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
  } else {
    v_Var->i_Indicator[0] = OCI_IND_NULL;
  }
  return (PyObject*) v_Var;
}

