//-----------------------------------------------------------------------------
// ExternalLobVar.c
//   Defines the routines for handling LOB variables external to this module.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// external LOB type
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_HEAD
  udt_LobVar *i_Var;
  unsigned i_Pos;
} udt_ExternalLobVar;


//-----------------------------------------------------------------------------
// Declaration of external LOB variable functions.
//-----------------------------------------------------------------------------
static void ExternalLobVar_Free(udt_ExternalLobVar*);
static PyObject *ExternalLobVar_GetAttr(udt_ExternalLobVar*, PyObject*);
static PyObject *ExternalLobVar_Size(udt_ExternalLobVar*, PyObject*);
static PyObject *ExternalLobVar_Read(udt_ExternalLobVar*, PyObject*);


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_ExternalLobVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "ExternalLobVar",			// type name string
  sizeof(udt_ExternalLobVar),		// basic size (memory allocated)
  0,					// item size (variable object)
  (destructor) ExternalLobVar_Free,	// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) ExternalLobVar_GetAttr,
					// getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// declaration of methods for Python type "ExternalLOBVar"
//-----------------------------------------------------------------------------
static PyMethodDef g_ExternalLobVarMethods[] = {
  { "size", (PyCFunction) ExternalLobVar_Size, METH_VARARGS },
  { "read", (PyCFunction) ExternalLobVar_Read, METH_VARARGS },
  { NULL, NULL }
};


//-----------------------------------------------------------------------------
// ExternalLobVar_New()
//   Create a new external LOB variable.
//-----------------------------------------------------------------------------
PyObject *ExternalLobVar_New(
  udt_LobVar *a_Var,			// variable to encapsulate
  unsigned a_Pos)			// position in array to encapsulate
{
  udt_ExternalLobVar *v_Var;

  v_Var = PyObject_NEW(udt_ExternalLobVar, &g_ExternalLobVarType);
  if (!v_Var)
    return NULL;
  v_Var->i_Pos = a_Pos;
  v_Var->i_Var = a_Var;
  Py_INCREF(v_Var->i_Var);

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Free()
//   Free an external LOB variable.
//-----------------------------------------------------------------------------
static void ExternalLobVar_Free(
  udt_ExternalLobVar *a_Var)		// variable to free
{
  Py_DECREF(a_Var->i_Var);
  PyMem_DEL(a_Var);
}


//-----------------------------------------------------------------------------
// ExternalLobVar_GetAttr()
//   Retrieve an attribute on the external LOB variable object.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_GetAttr(
  udt_ExternalLobVar *a_Var,		// cursor object
  PyObject *a_Name)			// name of attribute
{
  return Py_FindMethod(g_ExternalLobVarMethods, (PyObject*) a_Var,
      PyString_AS_STRING(a_Name));
}


//-----------------------------------------------------------------------------
// ExternalLobVar_InternalSize()
//   Return the size of the LOB variable for internal comsumption.
//-----------------------------------------------------------------------------
int ExternalLobVar_InternalSize(
  udt_ExternalLobVar *a_Var)		// variable to return the size of
{
  sword v_Status;
  ub4 v_Length;

  v_Status = OCILobGetLength(a_Var->i_Var->i_Connection->i_Handle,
      g_ErrorHandle, a_Var->i_Var->i_Data[a_Var->i_Pos], &v_Length);
  if (CheckForError(v_Status) < 0)
    return -1;

  return v_Length;
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Size()
//   Return the size of the LOB variable.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Size(
  udt_ExternalLobVar *a_Var,		// variable to return the size of
  PyObject *a_Args)			// arguments
{
  int v_Length;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // return value
  v_Length = ExternalLobVar_InternalSize(a_Var);
  if (v_Length < 0)
    return NULL;
  return PyInt_FromLong(v_Length);
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Read()
//   Return a portion of the external LOB variable.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Read(
  udt_ExternalLobVar *a_Var,		// variable to return the size of
  PyObject *a_Args)			// arguments
{
  PyObject *v_Result;
  int v_Offset, v_Amount;
  sword v_Status;
  char *v_Value;
  ub4 v_Length;

  // offset and amount are expected, both optional
  v_Offset = v_Amount = -1;
  if (!PyArg_ParseTuple(a_Args, "|ii", &v_Offset, &v_Amount))
    return NULL;

  // modify the arguments
  if (v_Offset < 0)
    v_Offset = 1;
  if (v_Amount < 0) {
    v_Amount = ExternalLobVar_InternalSize(a_Var);
    if (v_Amount < 0)
      return NULL;
    v_Amount = v_Amount - v_Offset + 1;
    if (v_Amount <= 0)
      v_Amount = 1;
  }
  v_Length = v_Amount;

  // create a string for retrieving the value
  v_Value = PyMem_Malloc(v_Length);
  if (!v_Value)
    return PyErr_NoMemory();
  v_Status = OCILobRead(a_Var->i_Var->i_Connection->i_Handle,
      g_ErrorHandle, a_Var->i_Var->i_Data[a_Var->i_Pos], &v_Length, v_Offset,
      v_Value, v_Length, NULL, NULL, 0, 0); 
  if (CheckForError(v_Status) < 0)
    return NULL;

  // return the result
  v_Result = PyString_FromStringAndSize(v_Value, v_Length);
  PyMem_Free(v_Value);
  return v_Result;
}

