//-----------------------------------------------------------------------------
// LobVar.c
//   Defines the routines for handling LOB variables.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// LOB type
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_VAR_HEAD
  udt_Connection *i_Connection;
  ub4 i_MaxRows;
  ub2 i_DataType;
  sb2 *i_Indicator;
  OCILobLocator **i_Data;
} udt_LobVar;

#include "ExternalLobVar.c"

//-----------------------------------------------------------------------------
// Declaration of LOB variable functions.
//-----------------------------------------------------------------------------
static void LobVar_Free(udt_LobVar*);


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_LobVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "LobVar",				// type name string
  sizeof(udt_LobVar),			// basic size (memory allocated)
  sizeof(OCILobLocator**) + sizeof(sb2),
					// item size (variable object)
  (destructor) LobVar_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  0,					// getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// LobVar_New()
//   Create a new LOB variable.
//-----------------------------------------------------------------------------
PyObject *LobVar_New(
  udt_Connection *a_Connection,		// connection that is creating variable
  unsigned a_NumRows,			// number of rows to allocate
  ub2 a_DataType)			// data type
{
  udt_LobVar *v_Var;
  sword v_Status;
  int i;

  // create Python object
  v_Var = PyObject_NEW_VAR(udt_LobVar, &g_LobVarType, a_NumRows);
  if (!v_Var)
    return NULL;

  // initialize members
  v_Var->i_DataType = a_DataType;
  v_Var->i_MaxRows = a_NumRows;
  v_Var->i_Indicator = (sb2*) (((char*) v_Var) + sizeof(udt_LobVar));
  v_Var->i_Data = (OCILobLocator**) &(v_Var->i_Indicator[a_NumRows]);
  v_Var->i_Connection = a_Connection;
  Py_INCREF(v_Var->i_Connection);

  // initialize the LOB locators
  for (i = 0; i < a_NumRows; i++) {
    v_Status = OCIDescriptorAlloc(g_EnvHandle, (dvoid**) &v_Var->i_Data[i],
        OCI_DTYPE_LOB, 0, 0);
    if (CheckForError(v_Status) < 0)
      return NULL;
  }

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// LobVar_Free()
//   Free a LOB variable.
//-----------------------------------------------------------------------------
static void LobVar_Free(
  udt_LobVar *a_Var)			// variable to free
{
  int i;

  for (i = 0; i < a_Var->i_MaxRows; i++) {
    if (a_Var->i_Data[i])
      OCIDescriptorFree(a_Var->i_Data[i], OCI_DTYPE_LOB);
  }
  Py_DECREF(a_Var->i_Connection);
  PyMem_DEL(a_Var);
}


//-----------------------------------------------------------------------------
// LobVar_Define()
//   Define a LOB variable at the specified position in the cursor.
//-----------------------------------------------------------------------------
PyObject *LobVar_Define(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  int a_Position,			// position at which to perform define
  ub2 a_DataType)			// data type of LOB
{
  OCIDefine *v_Handle;
  udt_LobVar *v_Var;
  sword v_Status;

  // create a number variable
  v_Var = (udt_LobVar*) LobVar_New(a_Cursor->i_Connection,
      a_Cursor->i_AllocatedRows, a_DataType);
  if (!v_Var)
    return NULL;

  // perform the define
  v_Handle = 0;
  v_Status = OCIDefineByPos(a_Cursor->i_Handle, &v_Handle, g_ErrorHandle,
      a_Position, v_Var->i_Data, sizeof(OCINumber), v_Var->i_DataType,
      v_Var->i_Indicator, 0, 0, OCI_DEFAULT);
  if (CheckForError(v_Status) < 0) {
    Py_DECREF(v_Var);
    return NULL;
  }

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// LobVar_Value()
//   Returns the value for a LOB variable.
//-----------------------------------------------------------------------------
PyObject *LobVar_Value(
  udt_LobVar *a_Var,			// variable to determine value for
  int a_ArrayPos)			// array position
{
  if (a_Var->i_Indicator[a_ArrayPos] == OCI_IND_NULL) {
    Py_INCREF(Py_None);
    return Py_None;
  }
  return ExternalLobVar_New(a_Var, a_ArrayPos);
}

