//-----------------------------------------------------------------------------
// NumberVar.c
//   Defines the routines for handling numeric variables.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Number type
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_VAR_HEAD
  ub4 i_MaxRows;
  int i_IsInteger;
  sb2 *i_Indicator;
  OCINumber *i_Data;
} udt_NumberVar;


//-----------------------------------------------------------------------------
// Declaration of number variable functions.
//-----------------------------------------------------------------------------
static void NumberVar_Free(udt_NumberVar*);


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_NumberVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "NumberVar",				// type name string
  sizeof(udt_NumberVar),		// basic size (memory allocated)
  sizeof(OCINumber) + sizeof(sb2),	// item size (variable object)
  (destructor) NumberVar_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  0,					// getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// NumberVar_New()
//   Create a new number variable.
//-----------------------------------------------------------------------------
PyObject *NumberVar_New(
  unsigned a_NumRows,			// number of rows to allocate
  int a_IsInteger)			// is integer?
{
  udt_NumberVar *v_Var;

  // create Python object
  v_Var = PyObject_NEW_VAR(udt_NumberVar, &g_NumberVarType, a_NumRows);
  if (!v_Var)
    return NULL;

  // initialize members
  v_Var->i_MaxRows = a_NumRows;
  v_Var->i_IsInteger = a_IsInteger;
  v_Var->i_Indicator = (sb2*) (((char*) v_Var) + sizeof(udt_NumberVar));
  v_Var->i_Data = (OCINumber*) &(v_Var->i_Indicator[a_NumRows]);

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// NumberVar_Free()
//   Free a number variable.
//-----------------------------------------------------------------------------
static void NumberVar_Free(
  udt_NumberVar *a_Var)			// variable to free
{
  PyMem_DEL(a_Var);
}


//-----------------------------------------------------------------------------
// NumberVar_Define()
//   Define a number variable at the specified position in the cursor.
//-----------------------------------------------------------------------------
PyObject *NumberVar_Define(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  int a_Position,			// position at which to perform define
  OCIParam *a_Param)			// parameter descriptor
{
  OCIDefine *v_Handle;
  udt_NumberVar *v_Var;
  ub1 v_Precision;
  sword v_Status;
  sb1 v_Scale;

  // determine if number can be treated as an integer
  v_Status = OCIAttrGet(a_Param, OCI_DTYPE_PARAM, (dvoid*) &v_Scale, 0,
      OCI_ATTR_SCALE, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;
  v_Status = OCIAttrGet(a_Param, OCI_DTYPE_PARAM, (dvoid*) &v_Precision, 0,
      OCI_ATTR_PRECISION, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;

  // create a number variable
  v_Var = (udt_NumberVar*) NumberVar_New(a_Cursor->i_AllocatedRows,
      (v_Scale == 0 && v_Precision > 0 && v_Precision < 10));
  if (!v_Var)
    return NULL;

  // perform the define
  v_Handle = 0;
  v_Status = OCIDefineByPos(a_Cursor->i_Handle, &v_Handle, g_ErrorHandle,
      a_Position, v_Var->i_Data, sizeof(OCINumber),  SQLT_VNU,
      v_Var->i_Indicator, 0, 0, OCI_DEFAULT);
  if (CheckForError(v_Status) < 0) {
    Py_DECREF(v_Var);
    return NULL;
  }

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// NumberVar_BindToCursor()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int NumberVar_BindToCursor(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  udt_NumberVar *a_Var)			// variable to bind
{
  OCIBind *v_BindHandle;
  sword v_Status;

  v_BindHandle = NULL;
  v_Status = OCIBindByName(a_Cursor->i_Handle, &v_BindHandle, g_ErrorHandle,
      PyString_AS_STRING(a_Name), PyString_GET_SIZE(a_Name), a_Var->i_Data,
      sizeof(OCINumber), SQLT_VNU, a_Var->i_Indicator, 0, 0, 0, 0,
      OCI_DEFAULT);
  if (CheckForError(v_Status) < 0)
    return -1;

  return 0;
}


//-----------------------------------------------------------------------------
// NumberVar_SetValue()
//   Set the value of the number.
//-----------------------------------------------------------------------------
static int NumberVar_SetValue(
  udt_NumberVar *a_Var,			// variable to set
  int a_Pos,				// position to set
  PyObject *a_Value)			// value to set
{
  double v_DoubleValue;
  int v_IntegerValue;
  sword v_Status;

  if (a_Value == Py_None)
    a_Var->i_Indicator[a_Pos] = OCI_IND_NULL;
  else {
    a_Var->i_Indicator[a_Pos] = OCI_IND_NOTNULL;
    if (PyInt_Check(a_Value)) {
      a_Var->i_IsInteger = 1;
      v_IntegerValue = PyInt_AS_LONG(a_Value);
      v_Status = OCINumberFromInt(g_ErrorHandle, &v_IntegerValue,
          sizeof(int), OCI_NUMBER_SIGNED, &a_Var->i_Data[a_Pos]);
    } else {
      a_Var->i_IsInteger = 0;
      v_DoubleValue = PyFloat_AS_DOUBLE(a_Value);
      v_Status = OCINumberFromReal(g_ErrorHandle, &v_DoubleValue,
          sizeof(double), &a_Var->i_Data[a_Pos]);
    }
    if (CheckForError(v_Status) < 0)
      return -1;
  }
  return 0;
}


//-----------------------------------------------------------------------------
// NumberVar_Bind()
//   Bind a number to the variable name.
//-----------------------------------------------------------------------------
PyObject *NumberVar_Bind(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  PyObject *a_Value)			// value (null or integer/float types)
{
  udt_NumberVar *v_Var;

  v_Var = (udt_NumberVar*) NumberVar_New(1, 0);
  if (!v_Var)
    return NULL;
  if (a_Value) {
    if (NumberVar_SetValue(v_Var, 0, a_Value) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
    if (NumberVar_BindToCursor(a_Cursor, a_Name, v_Var) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
  } else {
    v_Var->i_Indicator[0] = OCI_IND_NULL;
  }
  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// NumberVar_Value()
//   Returns the value for a number variable.
//-----------------------------------------------------------------------------
PyObject *NumberVar_Value(
  udt_NumberVar *a_Var,			// variable to determine value for
  int a_ArrayPos)			// array position
{
  double v_DoubleValue;
  int v_IntegerValue;
  sword v_Status;

  if (a_Var->i_Indicator[a_ArrayPos] == OCI_IND_NULL) {
    Py_INCREF(Py_None);
    return Py_None;
  }
  if (a_Var->i_IsInteger) {
    v_Status = OCINumberToInt(g_ErrorHandle, &a_Var->i_Data[a_ArrayPos],
        sizeof(int), OCI_NUMBER_SIGNED, (dvoid*) &v_IntegerValue);
    if (CheckForError(v_Status) < 0)
      return NULL;
    return PyInt_FromLong(v_IntegerValue);
  }
  v_Status = OCINumberToReal(g_ErrorHandle, &a_Var->i_Data[a_ArrayPos],
      sizeof(double), (dvoid*) &v_DoubleValue);
  if (CheckForError(v_Status) < 0)
    return NULL;
  return PyFloat_FromDouble(v_DoubleValue);
}

//-----------------------------------------------------------------------------
// NumberVar_BindArrayToCursor()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int NumberVar_BindArrayToCursor(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  udt_NumberVar *a_Var)			// variable to bind
{
  OCIBind *v_BindHandle;
  sword v_Status;

  v_BindHandle = NULL;
  v_Status = OCIBindByName(a_Cursor->i_Handle, &v_BindHandle, g_ErrorHandle,
      PyString_AS_STRING(a_Name), PyString_GET_SIZE(a_Name), a_Var->i_Data,
      sizeof(OCINumber), SQLT_VNU, a_Var->i_Indicator, 0, 0, a_Var->i_MaxRows,
      &(a_Var->i_MaxRows), OCI_DEFAULT);
  if (CheckForError(v_Status) < 0)
    return -1;

  return 0;
}

//-----------------------------------------------------------------------------
// NumberVar_BindArray()
//   Bind a number to the variable name.
//-----------------------------------------------------------------------------
PyObject *NumberVar_BindArray(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  PyObject *a_Value)			// value (null or integer/float types)
{
  udt_NumberVar *v_Var;
  int i, v_Size;

  v_Size = PyList_GET_SIZE(a_Value);
  v_Var = (udt_NumberVar*) NumberVar_New(v_Size, 0);
  if (!v_Var)
    return NULL;
  if (a_Value) {
    for (i = 0; i < v_Size; i++) {
      if (NumberVar_SetValue(v_Var, i, PyList_GET_ITEM(a_Value, i)) < 0) {
        Py_DECREF(v_Var);
        return NULL;
      }
    }
    if (NumberVar_BindArrayToCursor(a_Cursor, a_Name, v_Var) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
  } else {
    v_Var->i_Indicator[0] = OCI_IND_NULL;
  }
  return (PyObject*) v_Var;
}

