//-----------------------------------------------------------------------------
// StringVar.c
//   Defines the routines specific to the string type.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// String type
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_VAR_HEAD
  ub4 i_MaxRows;
  unsigned i_MaxLength;
  sb2 *i_Indicator;
  ub2 *i_ReturnCode;
  ub2 *i_ActualLength;
  char *i_Data;
} udt_StringVar;


//-----------------------------------------------------------------------------
// Declaration of string variable functions.
//-----------------------------------------------------------------------------
static void StringVar_Free(udt_StringVar*);


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_StringVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "StringVar",				// type name string
  sizeof(udt_StringVar),		// basic size (memory allocated)
  1,					// item size (variable object)
  (destructor) StringVar_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  0,					// getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// StringVar_New()
//   Create a new string variable.
//-----------------------------------------------------------------------------
PyObject *StringVar_New(
  unsigned a_NumRows,			// number of rows to allocate
  ub2 a_MaxLength)			// maximum length of the string
{
  udt_StringVar *v_Var;
  int v_Size;

  // create Python object
  v_Size = (a_MaxLength + sizeof(ub2) + sizeof(ub2) + sizeof(sb2)) * a_NumRows;
  v_Var = PyObject_NEW_VAR(udt_StringVar, &g_StringVarType, v_Size);
  if (!v_Var)
    return NULL;

  // initialize members
  v_Var->i_MaxRows = a_NumRows;
  v_Var->i_MaxLength = a_MaxLength;
  v_Var->i_Indicator = (sb2*) (((char*) v_Var) + sizeof(udt_StringVar));
  v_Var->i_ReturnCode = (ub2*) &v_Var->i_Indicator[a_NumRows];
  v_Var->i_ActualLength = (ub2*) &v_Var->i_ReturnCode[a_NumRows];
  v_Var->i_Data = (char*) &v_Var->i_ActualLength[a_NumRows];

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// StringVar_Free()
//   Free a string variable.
//-----------------------------------------------------------------------------
static void StringVar_Free(
  udt_StringVar *a_Var)			// variable to free
{
  PyMem_DEL(a_Var);
}


//-----------------------------------------------------------------------------
// StringVar_Define()
//   Define a string variable at the specified position in the cursor.
//-----------------------------------------------------------------------------
PyObject *StringVar_Define(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  int a_Position,			// position at which to perform define
  OCIParam *a_Param)			// parameter descriptor
{
  OCIDefine *v_Handle;
  udt_StringVar *v_Var;
  ub2 v_MaxLength;
  sword v_Status;

  // determine the maximum length
  v_Status = OCIAttrGet(a_Param, OCI_DTYPE_PARAM, (dvoid*) &v_MaxLength,
      0, OCI_ATTR_DATA_SIZE, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;
  if (v_MaxLength == 0)
    v_MaxLength = 16384;

  // create a string variable
  v_Var = (udt_StringVar*) StringVar_New(a_Cursor->i_AllocatedRows,
      v_MaxLength);
  if (!v_Var)
    return NULL;

  // perform the define
  v_Handle = 0;
  v_Status = OCIDefineByPos(a_Cursor->i_Handle, &v_Handle, g_ErrorHandle,
      a_Position, v_Var->i_Data, v_MaxLength,  SQLT_CHR,
      v_Var->i_Indicator, v_Var->i_ActualLength, v_Var->i_ReturnCode,
      OCI_DEFAULT);
  if (CheckForError(v_Status) < 0) {
    Py_DECREF(v_Var);
    return NULL;
  }

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// StringVar_BindToCursor()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int StringVar_BindToCursor(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  udt_StringVar *a_Var)			// variable to bind
{
  OCIBind *v_BindHandle;
  sword v_Status;

  v_BindHandle = NULL;
  v_Status = OCIBindByName(a_Cursor->i_Handle, &v_BindHandle, g_ErrorHandle,
      PyString_AS_STRING(a_Name), PyString_GET_SIZE(a_Name), a_Var->i_Data,
      a_Var->i_MaxLength, SQLT_CHR, a_Var->i_Indicator, a_Var->i_ActualLength,
      a_Var->i_ReturnCode, 0, 0, OCI_DEFAULT);
  if (CheckForError(v_Status) < 0)
    return -1;

  return 0;
}


//-----------------------------------------------------------------------------
// StringVar_SetValue()
//   Bind a string to the variable name.
//-----------------------------------------------------------------------------
static int StringVar_SetValue(
  udt_StringVar *a_Var,			// variable to set
  int a_Pos,				// position to set
  PyObject *a_Value)			// value to set
{
  if (a_Value == Py_None)
    a_Var->i_Indicator[a_Pos] = OCI_IND_NULL;
  else {
    a_Var->i_Indicator[a_Pos] = OCI_IND_NOTNULL;
    a_Var->i_ActualLength[a_Pos] = PyString_GET_SIZE(a_Value);
    memcpy(a_Var->i_Data + a_Pos * a_Var->i_MaxLength,
        PyString_AS_STRING(a_Value), PyString_GET_SIZE(a_Value));
  }
  return 0;
}


//-----------------------------------------------------------------------------
// StringVar_Bind()
//   Create a new string variable of the given length and bind it to the given
// cursor.
//-----------------------------------------------------------------------------
PyObject *StringVar_Bind(
  udt_Cursor *a_Cursor,			// cursor to bind to
  int a_MaxLength,			// maximum length to setup
  PyObject *a_Name,			// name of bind variable
  PyObject *a_Value)			// value to bind
{
  udt_StringVar *v_Var;

  v_Var = (udt_StringVar*) StringVar_New(1, a_MaxLength);
  if (!v_Var)
    return NULL;
  if (a_Value) {
    if (StringVar_SetValue(v_Var, 0, a_Value) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
    if (StringVar_BindToCursor(a_Cursor, a_Name, v_Var) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
  } else {
    v_Var->i_Indicator[0] = OCI_IND_NULL;
    v_Var->i_ActualLength[0] = a_MaxLength;
  }
  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// StringVar_Value()
//   Returns the value for a string variable.
//-----------------------------------------------------------------------------
PyObject *StringVar_Value(
  udt_StringVar *a_Var,			// variable to determine value for
  int a_ArrayPos)			// array position
{
  if (a_Var->i_Indicator[a_ArrayPos] == OCI_IND_NULL) {
    Py_INCREF(Py_None);
    return Py_None;
  }
  if (a_Var->i_ReturnCode[a_ArrayPos] == 1406) {
    PyErr_SetString(g_WarningException, "column truncation");
    return NULL;
  }
  return PyString_FromStringAndSize(
      a_Var->i_Data + a_ArrayPos * a_Var->i_MaxLength,
      a_Var->i_ActualLength[a_ArrayPos]);
}

//-----------------------------------------------------------------------------
// StringVar_BindArrayToCursor()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int StringVar_BindArrayToCursor(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  udt_StringVar *a_Var)			// variable to bind
{
  OCIBind *v_BindHandle;
  sword v_Status;

  v_BindHandle = NULL;
  v_Status = OCIBindByName(a_Cursor->i_Handle, &v_BindHandle, g_ErrorHandle,
      PyString_AS_STRING(a_Name), PyString_GET_SIZE(a_Name), a_Var->i_Data,
      a_Var->i_MaxLength, SQLT_CHR, a_Var->i_Indicator, a_Var->i_ActualLength,
      a_Var->i_ReturnCode, a_Var->i_MaxRows, &(a_Var->i_MaxRows), OCI_DEFAULT);
  if (CheckForError(v_Status) < 0)
    return -1;


  return 0;
}

//-----------------------------------------------------------------------------
// StringVar_BindArray()
//   Bind a number to the variable name (test).
//-----------------------------------------------------------------------------
PyObject *StringVar_BindArray(
  udt_Cursor *a_Cursor,			// cursor to perform define for
  PyObject *a_Name,			// name of bind variable
  PyObject *a_Value)			// value (null or integer/float types)
{
  int i, v_Size, v_MaxLength, v_TempLength;
  udt_StringVar *v_Var;
  PyObject *v_Value;

  // determine the maximum size of the strings involved
  v_Size = PyList_GET_SIZE(a_Value);
  v_MaxLength = 1;
  for (i = 0; i < v_Size; i++) {
    v_Value = PyList_GET_ITEM(a_Value, i);
    if (PyString_Check(v_Value)) {
      v_TempLength = PyString_GET_SIZE(v_Value);
      if (v_TempLength > v_MaxLength)
        v_MaxLength = v_TempLength;
    }
  }

  v_Var = (udt_StringVar*) StringVar_New(v_Size, v_MaxLength);
  if (!v_Var)
    return NULL;
  if (a_Value) {
    for (i = 0; i < v_Size; i++) {
      if (StringVar_SetValue(v_Var, i, PyList_GET_ITEM(a_Value, i)) < 0) {
        Py_DECREF(v_Var);
        return NULL;
      }
    }
    if (StringVar_BindArrayToCursor(a_Cursor, a_Name, v_Var) < 0) {
      Py_DECREF(v_Var);
      return NULL;
    }
  } else {
    v_Var->i_Indicator[0] = OCI_IND_NULL;
  }
  return (PyObject*) v_Var;
}

