//-----------------------------------------------------------------------------
// Variable.c
//   Defines Python types for Oracle variables.
//-----------------------------------------------------------------------------

#include "StringVar.c"
#include "NumberVar.c"
#include "DateTimeVar.c"
#include "LobVar.c"

//-----------------------------------------------------------------------------
// Variable_Define()
//   Allocate a variable and define it for the given statement.
//-----------------------------------------------------------------------------
static PyObject *Variable_Define(
  udt_Cursor *a_Cursor,			// cursor to define for
  int a_Position)			// position to define
{
  OCIParam *v_Param;
  ub2 v_DataType;
  sword v_Status;

  // retrieve parameter descriptor
  v_Status = OCIParamGet(a_Cursor->i_Handle, OCI_HTYPE_STMT, g_ErrorHandle,
      (dvoid**) &v_Param, a_Position);
  if (CheckForError(v_Status) < 0)
    return NULL;

  // retrieve datatype of the parameter
  v_Status = OCIAttrGet(v_Param, OCI_DTYPE_PARAM, (dvoid*) &v_DataType,
      0, OCI_ATTR_DATA_TYPE, g_ErrorHandle);
  if (CheckForError(v_Status) < 0)
    return NULL;

  // create a variable of the correct type
  switch(v_DataType) {
    case SQLT_LNG:
    case SQLT_AFC:
    case SQLT_CHR:
      return StringVar_Define(a_Cursor, a_Position, v_Param);
    case SQLT_NUM:
      return NumberVar_Define(a_Cursor, a_Position, v_Param);
    case SQLT_DAT:
      return DateTimeVar_Define(a_Cursor, a_Position, v_Param);
    case SQLT_CLOB:
    case SQLT_BLOB:
      return LobVar_Define(a_Cursor, a_Position, v_DataType);
  }

  PyErr_SetString(g_NotSupportedErrorException,
      "unhandled data type (define)");
  return NULL;
}


//-----------------------------------------------------------------------------
// Variable_Bind()
//   Allocate a variable and bind it to the given statement.
//-----------------------------------------------------------------------------
static PyObject *Variable_Bind(
  udt_Cursor *a_Cursor,			// cursor to bind variable to
  PyObject *a_Name,			// name to bind to
  PyObject *a_Value)			// value to bind
{
  int v_MaxLength, i;
  PyObject *v_Value;

  // handle strings
  if (a_Value == Py_None || PyString_Check(a_Value)) {
    if (a_Value == Py_None)
      v_MaxLength = 1;
    else v_MaxLength = PyString_GET_SIZE(a_Value);
    return StringVar_Bind(a_Cursor, v_MaxLength, a_Name, a_Value);
  }

  // handle numbers
  if (PyInt_Check(a_Value) || PyFloat_Check(a_Value))
    return NumberVar_Bind(a_Cursor, a_Name, a_Value);

  // handle dates
  if (a_Value->ob_type == &g_ExternalDateTimeVarType)
    return DateTimeVar_Bind(a_Cursor, a_Name, a_Value);

  // handle arrays
  if (PyList_Check(a_Value)) {
    for (i = 0; i < PyList_GET_SIZE(a_Value); i++) {
      v_Value = PyList_GET_ITEM(a_Value, i);
      if (PyInt_Check(v_Value) || PyFloat_Check(v_Value))
        return NumberVar_BindArray(a_Cursor, a_Name, a_Value);
      if (v_Value->ob_type == &g_ExternalDateTimeVarType)
        return DateTimeVar_BindArray(a_Cursor, a_Name, a_Value);
    }
    return StringVar_BindArray(a_Cursor, a_Name, a_Value);
  }

  // invalid input -- raise exception
  PyErr_SetString(g_NotSupportedErrorException, "unhandled data type (bind)");
  return NULL;
}


//-----------------------------------------------------------------------------
// Variable_BindToCursor()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int Variable_BindToCursor(
  udt_Cursor *a_Cursor,			// cursor to bind variable to
  PyObject *a_Name,			// name to bind to
  PyObject *a_Var)			// variable to bind
{
  if (a_Var->ob_type == &g_StringVarType)
    return StringVar_BindToCursor(a_Cursor, a_Name, (udt_StringVar*) a_Var);
  if (a_Var->ob_type == &g_NumberVarType)
    return NumberVar_BindToCursor(a_Cursor, a_Name, (udt_NumberVar*) a_Var);
  if (a_Var->ob_type == &g_DateTimeVarType)
    return DateTimeVar_BindToCursor(a_Cursor, a_Name,
        (udt_DateTimeVar*) a_Var);

  // invalid input -- raise exception
  PyErr_SetString(g_NotSupportedErrorException,
      "unhandled data type (bind to cursor)");
  return -1;
}


//-----------------------------------------------------------------------------
// Variable_BindNoValue()
//   Allocate a variable and bind it to the given statement.
//-----------------------------------------------------------------------------
static PyObject *Variable_BindNoValue(
  udt_Cursor *a_Cursor,			// cursor to bind variable to
  PyObject *a_Value)			// value to bind
{
  if (PyInt_Check(a_Value))
    return StringVar_Bind(a_Cursor, PyInt_AS_LONG(a_Value), NULL, NULL);
  if (a_Value == (PyObject*) &g_NumberVarType)
    return NumberVar_Bind(a_Cursor, NULL, NULL);
  if (a_Value == (PyObject*) &g_DateTimeVarType)
    return DateTimeVar_Bind(a_Cursor, NULL, NULL);

  // invalid input -- raise exception
  PyErr_SetString(g_NotSupportedErrorException,
      "unhandled data type (bind no value)");
  return NULL;
}


//-----------------------------------------------------------------------------
// Variable_Value()
//   Allocate a variable and define it for the given statement.
//-----------------------------------------------------------------------------
static PyObject *Variable_Value(
  PyObject *a_Var,			// variable to determine value for
  int a_ArrayPos)			// array position
{
  if (a_Var->ob_type == &g_StringVarType)
    return StringVar_Value((udt_StringVar*) a_Var, a_ArrayPos);
  if (a_Var->ob_type == &g_NumberVarType)
    return NumberVar_Value((udt_NumberVar*) a_Var, a_ArrayPos);
  if (a_Var->ob_type == &g_DateTimeVarType)
    return DateTimeVar_Value((udt_DateTimeVar*) a_Var, a_ArrayPos);
  if (a_Var->ob_type == &g_LobVarType)
    return LobVar_Value((udt_LobVar*) a_Var, a_ArrayPos);
  PyErr_SetString(g_NotSupportedErrorException, "unhandled data type (value)");
  return NULL;
}


//-----------------------------------------------------------------------------
// Variable_SetValue()
//   Allocate a variable and define it for the given statement.
//-----------------------------------------------------------------------------
static int Variable_SetValue(
  PyObject *a_Var,			// variable to set
  PyObject *a_Value)			// value to set
{
  if (a_Var->ob_type == &g_StringVarType)
    return StringVar_SetValue((udt_StringVar*) a_Var, 0, a_Value);
  if (a_Var->ob_type == &g_NumberVarType)
    return NumberVar_SetValue((udt_NumberVar*) a_Var, 0, a_Value);
  if (a_Var->ob_type == &g_ExternalDateTimeVarType)
    return DateTimeVar_SetValue((udt_DateTimeVar*) a_Var, 0, a_Value);
  PyErr_SetString(g_NotSupportedErrorException,
      "unhandled data type (set value)");
  return -1;
}

