//-----------------------------------------------------------------------------
// cx_Oracle.c
//   Shared library for use by Python.
//-----------------------------------------------------------------------------

#include <Python.h>
#include <oci.h>

//-----------------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------------
static OCIEnv *g_EnvHandle = NULL;
static OCIError *g_ErrorHandle = NULL;
static char g_ErrorBuffer[1024];
static PyObject *g_WarningException = NULL;
static PyObject *g_ErrorException = NULL;
static PyObject *g_InterfaceErrorException = NULL;
static PyObject *g_DatabaseErrorException = NULL;
static PyObject *g_DataErrorException = NULL;
static PyObject *g_OperationalErrorException = NULL;
static PyObject *g_IntegrityErrorException = NULL;
static PyObject *g_InternalErrorException = NULL;
static PyObject *g_ProgrammingErrorException = NULL;
static PyObject *g_NotSupportedErrorException = NULL;


//-----------------------------------------------------------------------------
// RaiseOracleError()
//   Reads the error that was caused by the last Oracle statement and raise an
// exception for Python. At this point it is assumed that the Oracle
// environment is fully initialized.
//-----------------------------------------------------------------------------
static void RaiseOracleError(void)
{
  sb4 v_ErrorNumber;
  sword v_Status;

  v_Status = OCIErrorGet(
      (dvoid*) g_ErrorHandle,		// error handle
      1,				// record number
      (text*) 0,			// SQL state (not supported in 8.0)
      &v_ErrorNumber,			// error number
      (text*) g_ErrorBuffer,		// error buffer
      (ub4) sizeof(g_ErrorBuffer),	// size of buffer available
      OCI_HTYPE_ERROR);			// type of handle
  if (v_Status == OCI_SUCCESS)
    PyErr_SetString(g_DatabaseErrorException, g_ErrorBuffer);
  else PyErr_SetString(g_InternalErrorException, "No Oracle error?");
}


//-----------------------------------------------------------------------------
// CheckForError()
//   Check for an error in the last call and if an error has occurred, raise a
// Python exception.
//-----------------------------------------------------------------------------
static int CheckForError(
  sword a_Status)			// status of last call
{
  if (a_Status != OCI_SUCCESS && a_Status != OCI_SUCCESS_WITH_INFO) {
    RaiseOracleError();
    return -1;
  }
  return 0;
}


//-----------------------------------------------------------------------------
// SetException()
//   Create an exception and set it in the provided dictionary.
//-----------------------------------------------------------------------------
static int SetException(
  PyObject *a_Dictionary,		// dictionary to modify
  PyObject **a_Exception,		// exception to create
  char *a_Name,				// name of the exception
  PyObject *a_BaseException)		// exception to base exception on
{
  char v_Name[100];

  sprintf(v_Name, "cx_Oracle.%s", a_Name);
  *a_Exception = PyErr_NewException(v_Name, a_BaseException, NULL);
  if (!*a_Exception)
    return -1;
  return PyDict_SetItemString(a_Dictionary, a_Name, *a_Exception);
}


#include "Connection.c"


//-----------------------------------------------------------------------------
// InitializeEnvironment()
//   Initialize the Oracle environment, if it has not already been initialized.
// Returns 0 on success.
//-----------------------------------------------------------------------------
static int InitializeEnvironment(void)
{
  sword v_Status;

  // do not initialize OCI environment more than once
  if (g_EnvHandle)
    return 0;

  // initialize the OCI environment
  OCIInitialize(
    OCI_OBJECT,				// mode
    NULL,				// context (for callbacks)
    NULL,				// memory allocation callback
    NULL,				// memory reallocation callback
    NULL);				// memory deallocation callback

  // call the OCI function to initialize the environment
  v_Status = OCIEnvInit(
      (OCIEnv**) &g_EnvHandle,		// handle (OUT)
      (ub4) OCI_DEFAULT,		// mode
      (size_t) 0,			// extra memory to allocate
      (dvoid**) 0);			// pointer to extra memory allocated
  if (v_Status != OCI_SUCCESS) {
    PyErr_SetString(PyExc_RuntimeError,
        "Unable to acquire Oracle environment handle");
    return -1;
  }

  // acquire an error handle
  v_Status = OCIHandleAlloc(
      (dvoid*) g_EnvHandle,		// environment handle
      (dvoid**) &g_ErrorHandle,		// error handle (OUT)
      (ub4) OCI_HTYPE_ERROR,		// type of handle
      (size_t) 0,			// extra memory to allocate
      (dvoid**) 0);			// pointer to extra memory allocated
  if (v_Status != OCI_SUCCESS) {
    OCIHandleFree(g_EnvHandle, OCI_HTYPE_ENV);
    g_EnvHandle = NULL;
    PyErr_SetString(PyExc_RuntimeError,
        "Unable to acquire Oracle error handle");
    return -1;
  }

  return 0;
}


//-----------------------------------------------------------------------------
// Connect()
//   Attempt to connect to the database using the username, password and TNS
// entry.
//-----------------------------------------------------------------------------
static PyObject* Connect(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  const char *v_UserName, *v_Password, *v_TNSEntry;
  unsigned v_UserNameLen, v_PasswordLen, v_TNSEntryLen;

  // parse arguments
  v_Password = v_TNSEntry = NULL;
  v_PasswordLen = v_TNSEntryLen = 0;
  if (!PyArg_ParseTuple(a_Args, "s#|s#s#", &v_UserName, &v_UserNameLen,
      &v_Password, &v_PasswordLen, &v_TNSEntry, &v_TNSEntryLen))
    return NULL;

  // if password or TNS entry not filled in, parse the connect string
  if (!v_Password || !v_TNSEntry) {
    v_TNSEntry = strchr(v_UserName, '@');
    if (v_TNSEntry) {
      v_TNSEntry++;
      v_TNSEntryLen = strlen(v_TNSEntry);
      v_UserNameLen -= v_TNSEntryLen + 1;
    }
    v_Password = strchr(v_UserName, '/');
    if (v_Password) {
      v_Password++;
      v_PasswordLen = strlen(v_Password);
      if (v_TNSEntryLen > 0)
        v_PasswordLen -= v_TNSEntryLen + 1;
      v_UserNameLen -= v_PasswordLen + 1;
    }
  }

  // make sure the Oracle environment is initialized
  if (InitializeEnvironment() < 0)
    return NULL;

  // create new service object
  return Connection_New(v_UserName, v_UserNameLen, v_Password, v_PasswordLen,
      v_TNSEntry, v_TNSEntryLen);
}


//-----------------------------------------------------------------------------
// Date()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* Date(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  int v_Year, v_Month, v_Day;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "iii", &v_Year, &v_Month, &v_Day))
    return NULL;

  return ExternalDateTimeVar_New(v_Year, v_Month, v_Day, 0, 0, 0);
}


//-----------------------------------------------------------------------------
// Time()
//   Returns a time value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* Time(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  int v_Hour, v_Minute, v_Second;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "iii", &v_Hour, &v_Minute, &v_Second))
    return NULL;

  return ExternalDateTimeVar_New(0, 0, 0, v_Hour, v_Minute, v_Second);
}


//-----------------------------------------------------------------------------
// Timestamp()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* Timestamp(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  int v_Year, v_Month, v_Day, v_Hour, v_Minute, v_Second;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "iiiiii", &v_Year, &v_Month, &v_Day, &v_Hour,
      &v_Minute, &v_Second))
    return NULL;

  return ExternalDateTimeVar_New(v_Year, v_Month, v_Day, v_Hour, v_Minute,
      v_Second);
}


//-----------------------------------------------------------------------------
//   Declaration of methods supported by this module
//-----------------------------------------------------------------------------
static PyMethodDef g_ModuleMethods[] = {
  { "connect", Connect, METH_VARARGS },
  { "Date", Date, METH_VARARGS },
  { "Time", Time, METH_VARARGS },
  { "Timestamp", Timestamp, METH_VARARGS },
  { NULL, NULL }
};

//-----------------------------------------------------------------------------
// initcx_Oracle()
//   Initialization routine for the shared libary.
//-----------------------------------------------------------------------------
void initcx_Oracle(void)
{
  PyObject *v_Module, *v_Dictionary;

  // initialize types
  g_ConnectionType.ob_type = &PyType_Type;
  g_CursorType.ob_type = &PyType_Type;
  g_StringVarType.ob_type = &PyType_Type;
  g_NumberVarType.ob_type = &PyType_Type;
  g_DateTimeVarType.ob_type = &PyType_Type;
  g_LobVarType.ob_type = &PyType_Type;
  g_ExternalLobVarType.ob_type = &PyType_Type;
  g_ExternalDateTimeVarType.ob_type = &PyType_Type;

  // initialize module and retrieve the dictionary
  v_Module = Py_InitModule("cx_Oracle", g_ModuleMethods);
  if (!v_Module)
    return;
  v_Dictionary = PyModule_GetDict(v_Module);
  if (!v_Dictionary)
    return;

  // create exception object and add it to the dictionary
  if (SetException(v_Dictionary, &g_WarningException, "Warning",
      PyExc_StandardError) < 0)
    return;
  if (SetException(v_Dictionary, &g_ErrorException, "Error",
      PyExc_StandardError) < 0)
    return;
  if (SetException(v_Dictionary, &g_InterfaceErrorException, "InterfaceError",
      g_ErrorException) < 0)
    return;
  if (SetException(v_Dictionary, &g_DatabaseErrorException, "DatabaseError",
      g_ErrorException) < 0)
    return;
  if (SetException(v_Dictionary, &g_DataErrorException, "DataError",
      g_DatabaseErrorException) < 0)
    return;
  if (SetException(v_Dictionary, &g_OperationalErrorException,
      "OperationalError", g_DatabaseErrorException) < 0)
    return;
  if (SetException(v_Dictionary, &g_IntegrityErrorException, "IntegrityError",
      g_DatabaseErrorException) < 0)
    return;
  if (SetException(v_Dictionary, &g_InternalErrorException, "InternalError",
      g_DatabaseErrorException) < 0)
    return;
  if (SetException(v_Dictionary, &g_ProgrammingErrorException,
      "ProgrammingError", g_DatabaseErrorException) < 0)
    return;
  if (SetException(v_Dictionary, &g_NotSupportedErrorException,
      "NotSupportedError", g_DatabaseErrorException) < 0)
    return;

  // create the basic data types for setting input sizes
  if (PyDict_SetItemString(v_Dictionary, "STRING",
      (PyObject*) &g_StringVarType) < 0)
    return;
  if (PyDict_SetItemString(v_Dictionary, "NUMBER",
      (PyObject*) &g_NumberVarType) < 0)
    return;
  if (PyDict_SetItemString(v_Dictionary, "DATETIME",
      (PyObject*) &g_DateTimeVarType) < 0)
    return;
  if (PyDict_SetItemString(v_Dictionary, "BINARY",
      (PyObject*) &g_LobVarType) < 0)
    return;

  // create constants required by Python DB API 2.0
  if (PyDict_SetItemString(v_Dictionary, "apilevel",
      PyString_FromString("2.0")) < 0)
    return;
  if (PyDict_SetItemString(v_Dictionary, "threadsafety",
      PyInt_FromLong(0)) < 0)
    return;
  if (PyDict_SetItemString(v_Dictionary, "paramstyle",
      PyString_FromString("named")) < 0)
    return;

  // initialize OCI environment
  InitializeEnvironment();
}

